/****************************************************************************
 Title	:   HD44780U LCD library
 Author:    Peter Fleury <pfleury@gmx.ch>  http://jump.to/fleury
 File:	    $Id: lcd.c,v 1.13.2.2 2004/02/12 21:08:25 peter Exp $
 Software:  AVR-GCC 3.3 
 Target:    any AVR device, memory mapped mode only for AT90S4414/8515/Mega

 DESCRIPTION
       Basic routines for interfacing a HD44780U-based text lcd display

       Originally based on Volker Oth's lcd library,
       changed lcd_init(), added additional constants for lcd_command(),
       added 4-bit I/O mode, improved and optimized code.

 USAGE
       See the C include lcd.h file for a description of each function

 AK: major changes
       
*****************************************************************************/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "lcd.h"

#include "delay.h"


/* 
** constants/macros 
*/
#define PIN(x) (*(&x - 2))  /* address of data direction register of port x */
#define DDR(x) (*(&x - 1))  /* address of input register of port x          */


#define lcd_e_delay()   delay_us(1)
#define lcd_e_high()    LCD_E_PORT  |=  _BV(LCD_E_PIN);
#define lcd_e_low()     LCD_E_PORT  &= ~_BV(LCD_E_PIN);
#define lcd_e_toggle()  toggle_e()
#define lcd_rs_high()   LCD_RS_PORT |=  _BV(LCD_RS_PIN)
#define lcd_rs_low()    LCD_RS_PORT &= ~_BV(LCD_RS_PIN)
#ifdef LCD_RW_PIN
  #define lcd_rw_high()   LCD_RW_PORT |=  _BV(LCD_RW_PIN)
  #define lcd_rw_low()    LCD_RW_PORT &= ~_BV(LCD_RW_PIN)
  #define lcd_waitcmd(us)
#else
  #define lcd_rw_low()
  #define lcd_waitcmd(us) delay_us(us)
  #define lcd_waitbusy()
#endif

#if LCD_LINES == 1
#  define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_1LINE 
#else
#  define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_2LINES 
#endif

#if LCD_LINES == 1
  #define DDADDR(x,y) (LCD_START_LINE1+(x))
#elif LCD_LINES == 2
  #define DDADDR(x,y) (((y) ? LCD_START_LINE2 : LCD_START_LINE1) + (x))
#else
  static uint8_t PROGMEM _ddaddr[4] = { LCD_START_LINE1, LCD_START_LINE2,
					LCD_START_LINE3, LCD_START_LINE4 };
  #define DDADDR(x,y) (_ddaddr[y]+(x))
#endif

static uint8_t row, col;

/*
** local functions
*/



/* toggle Enable Pin to initiate write */
static void toggle_e(void)
{
    lcd_e_high();
    lcd_e_delay();
    lcd_e_low();
}

/* set data direction out */
static void ddir_out(void)
{
    DDR(LCD_DATA0_PORT) |= _BV(LCD_DATA0_PIN);
    DDR(LCD_DATA1_PORT) |= _BV(LCD_DATA1_PIN);
    DDR(LCD_DATA2_PORT) |= _BV(LCD_DATA2_PIN);
    DDR(LCD_DATA3_PORT) |= _BV(LCD_DATA3_PIN);
}

/* set data direction in */
static void ddir_in(void)
{
    DDR(LCD_DATA0_PORT) &= ~_BV(LCD_DATA0_PIN);
    DDR(LCD_DATA1_PORT) &= ~_BV(LCD_DATA1_PIN);
    DDR(LCD_DATA2_PORT) &= ~_BV(LCD_DATA2_PIN);
    DDR(LCD_DATA3_PORT) &= ~_BV(LCD_DATA3_PIN);
}


/*************************************************************************
Low-level function to write byte to LCD controller
Input:    data   byte to write to LCD
          rs     1: write data    
                 0: write instruction
Returns:  none
*************************************************************************/
static void lcd_write4(uint8_t data)
{
    LCD_DATA3_PORT &= ~_BV(LCD_DATA3_PIN);
    LCD_DATA2_PORT &= ~_BV(LCD_DATA2_PIN);
    LCD_DATA1_PORT &= ~_BV(LCD_DATA1_PIN);
    LCD_DATA0_PORT &= ~_BV(LCD_DATA0_PIN);
    if (data & 0x08) LCD_DATA3_PORT |= _BV(LCD_DATA3_PIN);
    if (data & 0x04) LCD_DATA2_PORT |= _BV(LCD_DATA2_PIN);
    if (data & 0x02) LCD_DATA1_PORT |= _BV(LCD_DATA1_PIN);
    if (data & 0x01) LCD_DATA0_PORT |= _BV(LCD_DATA0_PIN);
    lcd_e_toggle();
}
static void lcd_write(uint8_t data, uint8_t rs) 
{
    if (rs) {   /* write data        (RS=1, RW=0) */
       lcd_rs_high();
    } else {    /* write instruction (RS=0, RW=0) */
       lcd_rs_low();
    }
    lcd_rw_low();

    /* configure data pins as output */
    ddir_out();
        
    /* output high nibble first */
    lcd_write4(data >> 4);
    lcd_write4(data);
        
    /* all data pins high (inactive) */
    ddir_in();
}


#ifdef LCD_RW_PIN

/*************************************************************************
Low-level function to read byte from LCD controller
Input:    rs     1: read data    
                 0: read busy flag / address counter
Returns:  byte read from LCD controller
*************************************************************************/
static uint8_t lcd_read4(void)
{
    uint8_t data;

    lcd_e_high();
    lcd_e_delay();        
    data = 0;
    if (PIN(LCD_DATA0_PORT) & _BV(LCD_DATA0_PIN)) data |= 0x01;
    if (PIN(LCD_DATA1_PORT) & _BV(LCD_DATA1_PIN)) data |= 0x02;
    if (PIN(LCD_DATA2_PORT) & _BV(LCD_DATA2_PIN)) data |= 0x04;
    if (PIN(LCD_DATA3_PORT) & _BV(LCD_DATA3_PIN)) data |= 0x08;
    lcd_e_low();
    return data;
}
static uint8_t lcd_read(uint8_t rs) 
{
    uint8_t data;
    
    if (rs)
        lcd_rs_high();                       /* RS=1: read data      */
    else
        lcd_rs_low();                        /* RS=0: read busy flag */
    lcd_rw_high();                           /* RW=1  read mode      */
    
    /* configure data pins as input */
    ddir_in();
                
    /* read high nibble first */
    data = lcd_read4() << 4;
    lcd_e_delay();
    data |= lcd_read4();

    return data;
}

/*************************************************************************
loops while lcd is busy
*************************************************************************/
static void lcd_waitbusy(void)

{
    /* wait until busy flag is cleared */
    while (lcd_read(0) & (1<<LCD_BUSY)) {
	;
    }

}/* lcd_waitbusy */

#endif /* LCD_RW_PIN */


/*
** PUBLIC FUNCTIONS 
*/

/*************************************************************************
Send LCD controller instruction command
Input:   instruction to send to LCD controller, see HD44780 data sheet
Returns: none
*************************************************************************/
void lcd_command(uint8_t cmd)
{
    lcd_waitbusy();
    lcd_write(cmd, 0);
    lcd_waitcmd(50);
}


/*************************************************************************
Set cursor to specified position
Input:    x  horizontal position  (0: left most position)
          y  vertical position    (0: first line)
Returns:  none
*************************************************************************/
void lcd_gotoxy(uint8_t x, uint8_t y)
{
    row = y;
    col = x;
    lcd_command((1<<LCD_DDRAM) + DDADDR(x,y));
}/* lcd_gotoxy */


/*************************************************************************
Clear display and set cursor to home position
*************************************************************************/
void lcd_clrscr(void)
{
    row = col = 0;
    lcd_command(1<<LCD_CLR);
    lcd_waitcmd(2000);
}


/*************************************************************************
Set cursor to home position
*************************************************************************/
void lcd_home(void)
{
    row = col = 0;
    lcd_command(1<<LCD_HOME);
    lcd_waitcmd(2000);
}


/*************************************************************************
Display character at current cursor position 
Input:    character to be displayed                                       
Returns:  none
*************************************************************************/
void lcd_putc(char c)
{
    int8_t i;
    const char *p;
    static char umlaut[] PROGMEM = {
	'', 0xE1,
	'', 0xEF,
	'', 0xF5,
	'', 0xE1,
	'', 0xEF,
	'', 0xF5,
	'', 0xE2,
	'', 0xDF,
    };

    switch (c) {

    case '\f':
	lcd_clrscr();
	break;

    case '\n':
	while (col < LCD_LINE_LENGTH)
	    lcd_putc(' ');
	lcd_gotoxy(0, (row + 1) % LCD_LINES);
	break;

    case '\r':
	lcd_gotoxy(0, row);
	break;

    default:
#if LCD_WRAP_LINES
	if (col >= LCL_DISP_LENGTH)
	    lcd_gotoxy(0, (row+1) % LCD_LINES);
#endif
	p = umlaut;
	for (i = 0; i < 8; ++i) {
	    if (c == pgm_read_byte(p))
		c = pgm_read_byte(p+1);
	    p += 2;
	}
	++col;
	lcd_waitbusy();
        lcd_write(c, 1);
	lcd_waitcmd(50);
    }

}/* lcd_putc */


/*************************************************************************
Display string without auto linefeed 
Input:    string to be displayed
Returns:  none
*************************************************************************/
void lcd_puts(const char *s)
/* print string on lcd (no auto linefeed) */
{
    register char c;

    while ((c = *s++) != '\0') {
        lcd_putc(c);
    }

}/* lcd_puts */


/*************************************************************************
Display string from program memory without auto linefeed 
Input:     string from program memory be be displayed                                        
Returns:   none
*************************************************************************/
void lcd_puts_p(const char *progmem_s)
/* print string from program memory on lcd (no auto linefeed) */
{
    register char c;

    while ((c = PRG_RDB(progmem_s++)) != '\0') {
        lcd_putc(c);
    }

}/* lcd_puts_p */


/*************************************************************************
Initialize display and select type of cursor 
Input:    dispAttr LCD_DISP_OFF            display off
                   LCD_DISP_ON             display on, cursor off
                   LCD_DISP_ON_CURSOR      display on, cursor on
                   LCD_DISP_CURSOR_BLINK   display on, cursor on flashing
Returns:  none
*************************************************************************/
void lcd_init(uint8_t dispAttr)
{
    /*
     *  Initialize LCD to 4 bit I/O mode
     */
     
    /* configure all port bits as output (LCD data and control lines on different ports */
    DDR(LCD_E_PORT)     |= _BV(LCD_E_PIN);
    DDR(LCD_RS_PORT)    |= _BV(LCD_RS_PIN);
#ifdef LCD_RW_PIN
    DDR(LCD_RW_PORT)    |= _BV(LCD_RW_PIN);
#endif
    ddir_out();
    
    /* initial write to lcd is 8bit */
    delay_ms(16);        /* wait 16ms or more after power-on       */
    LCD_DATA1_PORT |= _BV(LCD_DATA1_PIN);  // _BV(LCD_FUNCTION)>>4;
    LCD_DATA0_PORT |= _BV(LCD_DATA0_PIN);  // _BV(LCD_FUNCTION_8BIT)>>4;
    lcd_e_toggle();
    delay_ms(5);         /* delay, busy flag can't be checked here */
   
    /* repeat last command */ 
    lcd_e_toggle();      
    delay_us(100);       /* delay, busy flag can't be checked here */
    
    /* repeat last command a third time */
    lcd_e_toggle();      
    delay_us(100);       /* delay, busy flag can't be checked here */

    /* now configure for 4bit mode */
    LCD_DATA0_PORT &= ~_BV(LCD_DATA0_PIN);   // LCD_FUNCTION_4BIT_1LINE>>4
    lcd_e_toggle();
    delay_us(100);       /* some displays need this additional delay */
    
    /* from now the LCD only accepts 4 bit I/O, we can use lcd_command() */    
    lcd_command(LCD_FUNCTION_DEFAULT);      /* function set: display lines  */
    lcd_command(LCD_DISP_OFF);              /* display off                  */
    lcd_clrscr();                           /* display clear                */ 
    lcd_command(LCD_MODE_DEFAULT);          /* set entry mode               */
    lcd_command(dispAttr);                  /* display/cursor control       */

}/* lcd_init */
